from __future__ import annotations

import pytest

import env  # noqa: F401
import pybind11_tests.class_sh_trampoline_shared_ptr_cpp_arg as m


@pytest.mark.skipif("env.GRAALPY", reason="Cannot reliably trigger GC")
def test_shared_ptr_cpp_arg():
    import weakref

    class PyChild(m.SpBase):
        def is_base_used(self):
            return False

    tester = m.SpBaseTester()

    obj = PyChild()
    objref = weakref.ref(obj)

    # Pass the last python reference to the C++ function
    tester.set_object(obj)
    del obj
    pytest.gc_collect()

    # python reference is still around since C++ has it now
    assert objref() is not None
    assert tester.is_base_used() is False
    assert tester.obj.is_base_used() is False
    assert tester.get_object() is objref()


@pytest.mark.skipif("env.GRAALPY", reason="Cannot reliably trigger GC")
def test_shared_ptr_cpp_prop():
    class PyChild(m.SpBase):
        def is_base_used(self):
            return False

    tester = m.SpBaseTester()

    # Set the last python reference as a property of the C++ object
    tester.obj = PyChild()
    pytest.gc_collect()

    # python reference is still around since C++ has it now
    assert tester.is_base_used() is False
    assert tester.has_python_instance() is True
    assert tester.obj.is_base_used() is False
    assert tester.obj.has_python_instance() is True


@pytest.mark.skipif("env.GRAALPY", reason="Cannot reliably trigger GC")
def test_shared_ptr_arg_identity():
    import weakref

    tester = m.SpBaseTester()

    obj = m.SpBase()
    objref = weakref.ref(obj)

    tester.set_object(obj)
    del obj
    pytest.gc_collect()

    # NOTE: the behavior below is DIFFERENT from PR #2839
    # python reference is gone because it is not an Alias instance
    assert objref() is None
    assert tester.has_python_instance() is False


@pytest.mark.skipif("env.GRAALPY", reason="Cannot reliably trigger GC")
def test_shared_ptr_alias_nonpython():
    tester = m.SpBaseTester()

    # C++ creates the object, a python instance shouldn't exist
    tester.set_nonpython_instance()
    assert tester.is_base_used() is True
    assert tester.has_instance() is True
    assert tester.has_python_instance() is False

    # Now a python instance exists
    cobj = tester.get_object()
    assert cobj.has_python_instance()
    assert tester.has_instance() is True
    assert tester.has_python_instance() is True

    # Now it's gone
    del cobj
    pytest.gc_collect()
    assert tester.has_instance() is True
    assert tester.has_python_instance() is False

    # When we pass it as an arg to a new tester the python instance should
    # disappear because it wasn't created with an alias
    new_tester = m.SpBaseTester()

    cobj = tester.get_object()
    assert cobj.has_python_instance()

    new_tester.set_object(cobj)
    assert tester.has_python_instance() is True
    assert new_tester.has_python_instance() is True

    del cobj
    pytest.gc_collect()

    # Gone!
    assert tester.has_instance() is True
    assert tester.has_python_instance() is False
    assert new_tester.has_instance() is True
    assert new_tester.has_python_instance() is False


@pytest.mark.skipif("env.GRAALPY", reason="Cannot reliably trigger GC")
def test_shared_ptr_goaway():
    import weakref

    tester = m.SpGoAwayTester()

    obj = m.SpGoAway()
    objref = weakref.ref(obj)

    assert tester.obj is None

    tester.obj = obj
    del obj
    pytest.gc_collect()

    # python reference is no longer around
    assert objref() is None
    # C++ reference is still around
    assert tester.obj is not None


def test_infinite():
    tester = m.SpBaseTester()
    while True:
        tester.set_object(m.SpBase())
        break  # Comment out for manual leak checking (use `top` command).


@pytest.mark.parametrize(
    "pass_through_func", [m.pass_through_shd_ptr, m.pass_through_shd_ptr_release_gil]
)
def test_std_make_shared_factory(pass_through_func):
    class PyChild(m.SpBase):
        def __init__(self):
            super().__init__(0)

    obj = PyChild()
    while True:
        assert pass_through_func(obj) is obj
        break  # Comment out for manual leak checking (use `top` command).



#!/usr/bin/env python3
"""
Generate themed (light/dark) versions of constraint SVGs for KiCad DRC Rule Editor.

Color palette from https://dev-docs.kicad.org/en/rules-guidelines/icon-design/
"""

import os
import re
import shutil
from pathlib import Path

# Color transformations from light to dark theme (case-insensitive)
LIGHT_TO_DARK_COLORS = {
    '#545454': '#ded3dd',  # Primary Gray stroke
    '#606060': '#606060',  # Dense fills - stays same
    '#b9b9b9': '#e0e0e0',  # Accent Grey
    '#1a81c4': '#42b8eb',  # Primary Blue
    '#39b4ea': '#1a81c4',  # Accent Blue (swap)
    '#bf2641': '#f2647e',  # Primary Red
    '#f3f3f3': '#545454',  # Area Fill
    '#dcdcdc': '#545454',  # Light boundary lines
    '#000000': '#ded3dd',  # Black strokes become light
    '#000fe5': '#42b8eb',  # Blue arrows
}

# Files to process - mapping from source name to output name
# Format: 'source_filename.svg': 'output-name.svg'
CONSTRAINT_FILES = {
    # Specialized panels
    'absolute-lenght-2.svg': 'absolute-length-2.svg',
    'allowed-orientation.svg': 'allowed-orientation.svg',
    'vias-under-SMD.svg': 'vias-under-smd.svg',
    'minimum_text_height_and_thickness.svg': 'minimum-text-height-and-thickness.svg',
    'permitted-layers.svg': 'permitted-layers.svg',
    'via-style.svg': 'via-style.svg',

    # Numeric input constraints
    'basic-clearance.svg': 'basic-clearance.svg',
    'board-outline-clearance.svg': 'board-outline-clearance.svg',
    'clearance.svg': 'clearance.svg',
    'minimum_clearance.svg': 'minimum_clearance.svg',
    'minimum_connection_width.svg': 'minimum-connection-width.svg',
    'minimum_track_width.svg': 'minimum-track-width.svg',
    'copper_to_edge_clearance.svg': 'copper_to_edge_clearance.svg',
    'copper_to_hole_clearance.svg': 'copper-to-hole-clearance.svg',
    'creepage-distance.svg': 'creepage_distance.svg',
    'minimum_annular_width.svg': 'minimum-annular-width.svg',
    'minimum_drill_size.svg': 'minimum-drill-size.svg',
    'hole-to-hole-distance.svg': 'hole-to-hole-distance.svg',
    'minimum_via_diameter.svg': 'minimum-via-diameter.svg',
    'silk-to-soldermask-clearance.svg': 'silk-to-soldermask-clearance.svg',
    'silk-to-silk-clearance.svg': 'silk-to-silk-clearance.svg',
    'minimum-soldermask-sliver.svg': 'minimum-soldermask-sliver.svg',
    'minimum-soldermask-silver.svg': 'minimum-soldermask-silver.svg',
    'soldermask-expansion.svg': 'soldermask-expansion.svg',
    'solderpaste-expansion.svg': 'solderpaste-expansion.svg',
    'maximum_allowed_deviation.svg': 'maximum-allowed-deviation.svg',
    'matched-lenght-diff-pair.svg': 'matched-length-diff-pair.svg',
    "matched-length-diff-pair-v2.svg": "matched-length-diff-pair-v2.svg",
    'routing-diff-pair.svg': 'routing-diff-pair.svg',
    'routing-width.svg': 'routing-width.svg',
    'courtyard-clearance.svg': 'courtyard-clearance.svg',
}


def transform_colors_to_dark(svg_content: str) -> str:
    """Transform light theme colors to dark theme colors."""
    result = svg_content

    for light_color, dark_color in LIGHT_TO_DARK_COLORS.items():
        # Case-insensitive replacement for hex colors
        pattern = re.compile(re.escape(light_color), re.IGNORECASE)
        result = pattern.sub(dark_color, result)

    return result


def process_svg_files():
    """Process all constraint SVGs and create themed versions."""
    script_dir = Path(__file__).parent
    source_dir = script_dir / 'sources' / 'constraints'
    alias_source = source_dir / 'basic-clearance.svg'
    alias_target = source_dir / 'clearance.svg'
    if alias_source.exists() and not alias_target.exists():
        shutil.copy(alias_source, alias_target)
    light_dir = script_dir / 'sources' / 'light' / 'constraints'
    dark_dir = script_dir / 'sources' / 'dark' / 'constraints'

    # Create output directories if they don't exist
    light_dir.mkdir(parents=True, exist_ok=True)
    dark_dir.mkdir(parents=True, exist_ok=True)

    processed = 0
    skipped = 0

    for source_name, output_name in CONSTRAINT_FILES.items():
        source_path = source_dir / source_name

        if not source_path.exists():
            print(f"WARNING: Source file not found: {source_name}")
            skipped += 1
            continue

        # Read source SVG
        with open(source_path, 'r', encoding='utf-8') as f:
            svg_content = f.read()

        # Write light version (copy as-is, since originals use light theme colors)
        light_path = light_dir / output_name
        with open(light_path, 'w', encoding='utf-8') as f:
            f.write(svg_content)
        print(f"Created light: {output_name}")

        # Transform and write dark version
        dark_content = transform_colors_to_dark(svg_content)
        dark_path = dark_dir / output_name
        with open(dark_path, 'w', encoding='utf-8') as f:
            f.write(dark_content)
        print(f"Created dark:  {output_name}")

        processed += 1

    print(f"\nProcessed {processed} files, skipped {skipped}")
    return processed, skipped


if __name__ == '__main__':
    process_svg_files()



from __future__ import annotations

import contextlib
import os
import pickle
import sys

import pytest

# 3.14.0b3+, though sys.implementation.supports_isolated_interpreters is being added in b4
# Can be simplified when we drop support for the first three betas
CONCURRENT_INTERPRETERS_SUPPORT = (
    sys.version_info >= (3, 14)
    and (
        sys.version_info != (3, 14, 0, "beta", 1)
        and sys.version_info != (3, 14, 0, "beta", 2)
    )
    and (
        sys.version_info == (3, 14, 0, "beta", 3)
        or sys.implementation.supports_isolated_interpreters
    )
)


def get_interpreters(*, modern: bool):
    if modern and CONCURRENT_INTERPRETERS_SUPPORT:
        from concurrent import interpreters

        def create():
            return contextlib.closing(interpreters.create())

        def run_string(
            interp: interpreters.Interpreter,
            code: str,
            *,
            shared: dict[str, object] | None = None,
        ) -> Exception | None:
            if shared:
                interp.prepare_main(**shared)
            try:
                interp.exec(code)
                return None
            except interpreters.ExecutionFailed as err:
                return err

        return run_string, create

    if sys.version_info >= (3, 12):
        interpreters = pytest.importorskip(
            "_interpreters" if sys.version_info >= (3, 13) else "_xxsubinterpreters"
        )

        @contextlib.contextmanager
        def create(config: str = ""):
            try:
                if config:
                    interp = interpreters.create(config)
                else:
                    interp = interpreters.create()
            except TypeError:
                pytest.skip(f"interpreters module needs to support {config} config")

            try:
                yield interp
            finally:
                interpreters.destroy(interp)

        def run_string(
            interp: int, code: str, shared: dict[str, object] | None = None
        ) -> Exception | None:
            kwargs = {"shared": shared} if shared else {}
            return interpreters.run_string(interp, code, **kwargs)

        return run_string, create

    pytest.skip("Test requires the interpreters stdlib module")


@pytest.mark.skipif(
    sys.platform.startswith("emscripten"), reason="Requires loadable modules"
)
def test_independent_subinterpreters():
    """Makes sure the internals object differs across independent subinterpreters"""

    sys.path.append(".")

    run_string, create = get_interpreters(modern=True)

    m = pytest.importorskip("mod_per_interpreter_gil")

    if not m.defined_PYBIND11_HAS_SUBINTERPRETER_SUPPORT:
        pytest.skip("Does not have subinterpreter support compiled in")

    code = """
import mod_per_interpreter_gil as m
import pickle
with open(pipeo, 'wb') as f:
    pickle.dump(m.internals_at(), f)
"""

    with create() as interp1, create() as interp2:
        try:
            res0 = run_string(interp1, "import mod_shared_interpreter_gil")
            if res0 is not None:
                res0 = str(res0)
        except Exception as e:
            res0 = str(e)

        pipei, pipeo = os.pipe()
        run_string(interp1, code, shared={"pipeo": pipeo})
        with open(pipei, "rb") as f:
            res1 = pickle.load(f)

        pipei, pipeo = os.pipe()
        run_string(interp2, code, shared={"pipeo": pipeo})
        with open(pipei, "rb") as f:
            res2 = pickle.load(f)

    assert "does not support loading in subinterpreters" in res0, (
        "cannot use shared_gil in a default subinterpreter"
    )
    assert res1 != m.internals_at(), "internals should differ from main interpreter"
    assert res2 != m.internals_at(), "internals should differ from main interpreter"
    assert res1 != res2, "internals should differ between interpreters"


@pytest.mark.skipif(
    sys.platform.startswith("emscripten"), reason="Requires loadable modules"
)
@pytest.mark.skipif(not CONCURRENT_INTERPRETERS_SUPPORT, reason="Requires 3.14.0b3+")
def test_independent_subinterpreters_modern():
    """Makes sure the internals object differs across independent subinterpreters. Modern (3.14+) syntax."""

    sys.path.append(".")

    m = pytest.importorskip("mod_per_interpreter_gil")

    if not m.defined_PYBIND11_HAS_SUBINTERPRETER_SUPPORT:
        pytest.skip("Does not have subinterpreter support compiled in")

    from concurrent import interpreters

    code = """
import mod_per_interpreter_gil as m

values.put_nowait(m.internals_at())
"""

    with contextlib.closing(interpreters.create()) as interp1, contextlib.closing(
        interpreters.create()
    ) as interp2:
        with pytest.raises(
            interpreters.ExecutionFailed,
            match="does not support loading in subinterpreters",
        ):
            interp1.exec("import mod_shared_interpreter_gil")

        values = interpreters.create_queue()
        interp1.prepare_main(values=values)
        interp1.exec(code)
        res1 = values.get_nowait()

        interp2.prepare_main(values=values)
        interp2.exec(code)
        res2 = values.get_nowait()

    assert res1 != m.internals_at(), "internals should differ from main interpreter"
    assert res2 != m.internals_at(), "internals should differ from main interpreter"
    assert res1 != res2, "internals should differ between interpreters"


@pytest.mark.skipif(
    sys.platform.startswith("emscripten"), reason="Requires loadable modules"
)
def test_dependent_subinterpreters():
    """Makes sure the internals object differs across subinterpreters"""

    sys.path.append(".")

    run_string, create = get_interpreters(modern=False)

    m = pytest.importorskip("mod_shared_interpreter_gil")

    if not m.defined_PYBIND11_HAS_SUBINTERPRETER_SUPPORT:
        pytest.skip("Does not have subinterpreter support compiled in")

    code = """
import mod_shared_interpreter_gil as m
import pickle
with open(pipeo, 'wb') as f:
    pickle.dump(m.internals_at(), f)
"""

    with create("legacy") as interp1:
        pipei, pipeo = os.pipe()
        run_string(interp1, code, shared={"pipeo": pipeo})
        with open(pipei, "rb") as f:
            res1 = pickle.load(f)

    assert res1 != m.internals_at(), "internals should differ from main interpreter"



#!/usr/bin/env python

from __future__ import print_function

import sys
from pcbnew import *

# A helper function to convert a UTF8 string for python2 or python3
def fromUTF8Text( afilename ):
    if sys.version_info < (3, 0):
        return afilename.encode()
    else:
        return afilename

filename=sys.argv[1]

pcb = LoadBoard(filename)

ToUnits = ToMM
FromUnits = FromMM
#ToUnits=ToMils
#FromUnits=FromMils

print("List vias:")

for item in pcb.GetTracks():
    if type(item) is PCB_VIA:
        pos = item.GetPosition()
        drill = item.GetDrillValue()
        width = item.GetWidth()
        print(" * Via:   %s - %f/%f " % (ToUnits(pos), ToUnits(drill), ToUnits(width)))

    elif type(item) is PCB_TRACK:
        start = item.GetStart()
        end = item.GetEnd()
        width = item.GetWidth()
        print(" * Track: %s to %s, width %f" % (ToUnits(start), ToUnits(end), ToUnits(width)))

    elif type(item) is PCB_ARC:
        start = item.GetStart()
        end = item.GetEnd()
        width = item.GetWidth()
        print(" * Track arc: %s to %s, width %f" % (ToUnits(start), ToUnits(end), ToUnits(width)))

    else:
        print("Unknown type    %s" % type(item))

print("")
print("List drawings:")

for item in pcb.GetDrawings():
    if type(item) is PCB_TEXT:
        print("* Text:    '%s' at %s" % ( fromUTF8Text( item.GetText() ), ToUnits(item.GetPosition()) ) )
    elif type(item) is PCB_SHAPE:
        print( "* Drawing: %s" % fromUTF8Text( item.GetShapeStr() ) )
    else:
        print("* Drawing type: %s" % type(item))

print("")
print("List footprints:")

for footprint in pcb.Footprints():
    print("* Footprint: %s at %s" % ( fromUTF8Text( footprint.GetReference() ), ToUnits(footprint.GetPosition())))

print("")
print("Nets cnt: ", pcb.GetNetCount())
print("track w cnt:", len(pcb.GetTrackWidthList()))
print("via s cnt:", len(pcb.GetViasDimensionsList()))

print("")
print("List zones:", pcb.GetAreaCount())

for idx in range(0, pcb.GetAreaCount()):
    zone=pcb.GetArea(idx)
    print("zone:", idx, "priority:", zone.GetAssignedPriority(), "netname", fromUTF8Text( zone.GetNetname() ) )



#! /usr/bin/env python
"""
Generates a list of all codepoints in the Unicode range, skipping the surrogate range.

This is useful for generating a list of codepoints to test the newstroke tool.
"""

s = ""

# Skip the surrogate range
ranges = [
    (0x0000, 0xD800),
    (0xE000, 0xFFEF),
]

for rng in ranges:

    for i in range(rng[0], rng[1]):

        if i % 256 == 0:
            if i != 0:
                s += "\n"
            s += f"U+{i:04X} "
        elif i % 16 == 0:
            s += " "

        if i < 32:
            s += " "
        else:
            s += chr(i)

print(s)


from __future__ import annotations

import pybind11_tests.class_sh_trampoline_unique_ptr as m


class MyClass(m.Class):
    def foo(self):
        return 10 + self.get_val()

    def clone(self):
        cloned = MyClass()
        cloned.set_val(self.get_val() + 3)
        return cloned


def test_m_clone():
    obj = MyClass()
    while True:
        obj.set_val(5)
        obj = m.clone(obj)
        assert obj.get_val() == 5 + 3
        assert obj.foo() == 10 + 5 + 3
        return  # Comment out for manual leak checking (use `top` command).


def test_m_clone_and_foo():
    obj = MyClass()
    obj.set_val(7)
    while True:
        assert m.clone_and_foo(obj) == 10 + 7 + 3
        return  # Comment out for manual leak checking (use `top` command).



#
# This Aogram source code file is part of KiCad, a free EDA CAD application.
#
# Copyright The KiCad Developers, see AUTHORS.txt for contributors.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

from pathlib import Path
import pytest
import re

from conftest import KiTestFixture
import utils


def detect_indent(file_content: str) -> str:

    indent_regex = re.compile(r"^(?P<indent>\s+)", re.MULTILINE)
    match = indent_regex.search(file_content)
    if match:
        return match.group("indent")

    # Default to a tab if no symbols are found (or if the file is empty)
    return "\t"


def count_symbols_in_kicad_sym_file(file_path: Path) -> int:
    """
    Count the number of symbols in a .kicad_sym file by counting occurrences of "(symbol " at the start of lines.
    This is a simple heuristic that should work for normally-formatted .kicad_sym files.

    A more robust approach would be to parse the file as an S-expression and
    count the number of "symbol" entries.
    """
    with open(file_path, "r", encoding="utf-8") as f:
        data = f.read()

    indent = detect_indent(data)
    sym_regex = re.compile(rf'^{indent}\(symbol "', re.MULTILINE)
    return len(sym_regex.findall(data))


def assert_files_are_identical(file1: Path, file2: Path) -> None:
    with open(file1, "r", encoding="utf-8") as f1, open(
        file2, "r", encoding="utf-8"
    ) as f2:
        assert f1.read() == f2.read()


def get_sym_upgrade_cmd(input: str, output: str) -> list[str]:
    return [
        utils.kicad_cli(),
        "sym",
        "upgrade",
        "--force",
        str(input),
        "--output",
        str(output),
    ]


@pytest.mark.parametrize(
    "test_file,output_dir,num_syms",
    [
        # This is a v9-formatted library with 3 symbols
        (
            "cli/sym_lib_test/Amplifier_Video.v9.kicad_sym",
            "sym_lib_test",
            3,
        ),
    ],
)
def test_sym_upgrade_lib_pack_roundtrip(
    kitest: KiTestFixture,
    test_file: str,
    output_dir: str,
    num_syms: int,
) -> None:
    """
    Perform a roundtrip test of the "sym upgrade" command:
      - Unpack a .kicad_sym to a directory
      - Repack the directory to a .kicad_sym
      - Check the repacked .kicad_sym against the original
      - Upgradew the original .kicad_sym directly to another .kicad_sym and check that it's identical to the repacked one
      - Unpack the unpacked directory to another directory and check that the files are identical
    """

    input_file = kitest.get_data_file_path(test_file)
    output_path = kitest.get_output_path(f"cli/{output_dir}/")

    unpacked_lib_path = Path(output_path) / (
        input_file.stem + ".unpacked.kicad_symdir"
    )

    # Create output directory if it doesn't exist
    unpacked_lib_path.parent.mkdir(exist_ok=True)

    unpack_cmd = get_sym_upgrade_cmd(
        str(input_file),
        str(unpacked_lib_path) + "/",
    )
    stdout, stderr, exitcode = utils.run_and_capture(unpack_cmd)

    assert exitcode == 0
    assert stderr == ""
    assert stdout is not None

    # Check that the output is a directory
    assert unpacked_lib_path.is_dir()

    # Check that the number of symbols in the unpacked directory matches the expected count
    sym_files = list(unpacked_lib_path.glob("*.kicad_sym"))
    assert len(sym_files) == num_syms

    repacked_lib_path = Path(output_path) / (
        input_file.stem + ".repacked.kicad_sym"
    )

    if repacked_lib_path.exists():
        repacked_lib_path.unlink()

    repack_cmd = get_sym_upgrade_cmd(
        str(unpacked_lib_path),
        str(repacked_lib_path),
    )
    stdout, stderr, exitcode = utils.run_and_capture(repack_cmd)

    assert exitcode == 0
    assert stderr == ""
    assert stdout is not None

    # Check that the repacked file exists as a file
    assert repacked_lib_path.is_file()

    num_syms_in_repacked = count_symbols_in_kicad_sym_file(repacked_lib_path)
    assert num_syms_in_repacked == num_syms

    # Now, check that the packed file can be upgraded directly to another packed file
    packed_to_packed_output_path = Path(output_path) / (
        input_file.stem + ".packed_to_packed.kicad_sym"
    )

    if packed_to_packed_output_path.exists():
        packed_to_packed_output_path.unlink()

    packed_to_packed_command = get_sym_upgrade_cmd(
        str(input_file),
        str(packed_to_packed_output_path),
    )
    stdout, stderr, exitcode = utils.run_and_capture(packed_to_packed_command)

    assert exitcode == 0
    assert stderr == ""
    assert stdout is not None

    assert packed_to_packed_output_path.is_file()

    # This file should be identical to the repacked file
    assert_files_are_identical(repacked_lib_path, packed_to_packed_output_path)

    # Finally, check unpacking to unpacked lib again

    unpacked_to_unpacked_output_path = Path(output_path) / (
        input_file.stem + ".unpacked_to_unpacked.kicad_symdir"
    )

    unpacked_to_unpacked_cmd = get_sym_upgrade_cmd(
        str(unpacked_lib_path),
        str(unpacked_to_unpacked_output_path) + "/",
    )
    stdout, stderr, exitcode = utils.run_and_capture(unpacked_to_unpacked_cmd)

    assert exitcode == 0
    assert stderr == ""
    assert stdout is not None

    assert unpacked_to_unpacked_output_path.is_dir()

    pytest.skip("The following checks are currently failing: https://gitlab.com/kicad/code/kicad/-/issues/22988.")

    dir_a_files = sorted(unpacked_lib_path.glob("*.kicad_sym"))
    dir_b_files = sorted(unpacked_to_unpacked_output_path.glob("*.kicad_sym"))

    assert len(dir_a_files) == len(dir_b_files) == num_syms

    for file_a, file_b in zip(dir_a_files, dir_b_files):
        assert_files_are_identical(file_a, file_b)


@pytest.mark.parametrize(
    "input_lib,output_dir,num_units",
    [
        # This is a v9-formatted library with 3 symbols
        (
            "cli/sym_lib_test/Amplifier_Video.v9.kicad_sym",
            "sym_lib_test",
            4 + 1 + 1,
        ),
    ],
)
def test_sym_export_svg(
    kitest: KiTestFixture,
    input_lib: str,
    output_dir: str,
    num_units: int,
) -> None:
    """
    Perform an export of a footprint library to SVG.

    Note that much of the functionality of the "fp export svg" command is tested by the
    PCB "pcb export svg" tests, since they share the same underlying code.
    """

    input_file = kitest.get_data_file_path(input_lib)
    output_path = kitest.get_output_path(f"cli/{output_dir}/svg/")

    output_path.mkdir(parents=True, exist_ok=True)

    export_cmd = [
        utils.kicad_cli(),
        "sym",
        "export",
        "svg",
        str(input_file),
        "--output",
        str(output_path),
    ]

    stdout, stderr, exitcode = utils.run_and_capture(export_cmd)

    assert exitcode == 0
    assert stderr == ""
    assert stdout is not None

    # Check the number of footprints in the exported directory
    fp_files = list(Path(output_path).glob("*.svg"))
    assert len(fp_files) == num_units



#!/usr/bin/env python

# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright (C) 2016 CERN
# @author Maciej Suminski <maciej.suminski@cern.ch>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, you may find one here:
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# or you may search the http://www.gnu.org website for the version 2 license,
# or you may write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# Converts a bitmap font atlas description in XML format to data stored
# in a C structure.

import sys
import os

def convert(xml_file):
    txtfile = open( xml_file, "rt")
    rows = txtfile.readlines()
    output = open(os.path.splitext(xml_file)[0] + '_desc.c', 'w')
        
    # Header
    output.write(
"""
/* generated with fnt2struct.py, do not modify by hand */

static const struct bitmap_glyph {
    unsigned int x, y;
    unsigned int width, height;
    float x_off, y_off;
} bitmap_chars[] = {
""");

    last_id = 0
    fallback_line = '{ 0, 0, 0, 0, 0, 0 },\t\t/* %d (not defined) */\n'

    for row in rows[4:len(rows)-1:]:
        elements = row.split()
    
        cur_id = int(elements[1].split('=')[1])
        
        print('current id: %s' % cur_id)
                
        # Fill gaps for the missing characters
        while(cur_id > last_id):
            output.write(fallback_line % last_id)
            last_id = last_id + 1
            
        output.write('{ %d, %d, %d, %d, %ff, %ff },\t/* %d */\n' %
            (int(elements[2].split('=')[1]), int(elements[3].split('=')[1]),
            int(elements[4].split('=')[1]), int(elements[5].split('=')[1]),
            float(elements[6].split('=')[1]), float(elements[7].split('=')[1]),
            cur_id));
            
#        output.write('{ %d, %d, %d, %d, %d, %d },\t/* %d */\n' %
#                (int(char.attrib['x']), int(char.attrib['y']),
#                int(char.attrib['width']), int(char.attrib['height']),
#                int(char.attrib['xoffset']), int(char.attrib['yoffset']),
#                cur_id))
        last_id = cur_id + 1

    output.write('};\n')

    output.write('static const int bitmap_chars_count = %d;\n' % last_id)

#    try:
#        lineHeight = int(root.find('common').get('lineHeight'))
#    except:
#        print('Could not determine the font height')
#        lineHeight = -1

#    output.write('static const int bitmap_chars_height = %d;\n' % (lineHeight))

    output.close()

#----------------------------------------------------------------------
if __name__ == "__main__":
    argc = len(sys.argv)

    if(argc == 2):
        convert(sys.argv[1])
    else:
        print("usage: %s <xml-file>" % sys.argv[0])



import subprocess
import json
import re
import os
import sys

# --- Tree Generation Logic ---

KICAD_CLI = "kicad-cli"
ENV = os.environ.copy()
ENV["KICAD_RUN_FROM_BUILD_DIR"] = "1"

def get_help(args):
    cmd = [KICAD_CLI] + args + ["-h"]
    try:
        result = subprocess.run(cmd, capture_output=True, text=True, env=ENV)
        return result.stdout
    except Exception as e:
        return ""

def parse_help(help_text):
    subcommands = {}
    options = {}

    lines = help_text.split('\n')
    section = None

    for line in lines:
        line = line.strip()
        if not line:
            continue

        if line.startswith("Subcommands:"):
            section = "subcommands"
            continue
        elif line.startswith("Optional arguments:") or line.startswith("Arguments:"):
            section = "options"
            continue

        if section == "subcommands":
            match = re.match(r'^([a-z0-9_-]+)\s+(.*)$', line)
            if match:
                subcommands[match.group(1)] = match.group(2).strip()

        if section == "options":
            parts = re.split(r'\s{2,}', line, maxsplit=1)
            flags_part = parts[0]
            description = parts[1] if len(parts) > 1 else ""

            flags = re.findall(r'(-[a-zA-Z0-9]|--[a-zA-Z0-9_-]+)', flags_part)
            for flag in flags:
                options[flag] = description

    return subcommands, options

def explore(path):
    sys.stderr.write(f"Exploring {' '.join(path)}\n")
    help_text = get_help(path)
    subs_map, opts_map = parse_help(help_text)

    tree = {
        "options": opts_map,
        "subcommands": {},
        "subcommand_descriptions": subs_map
    }

    for sub in subs_map.keys():
        tree["subcommands"][sub] = explore(path + [sub])

    return tree

# --- Bash Completion Generation Logic ---

def generate_bash_completion(tree):
    script = []
    script.append("# kicad-cli bash completion")
    script.append("")
    script.append("_kicad_cli()")
    script.append("{")
    script.append("    local cur prev words cword")
    script.append("    if type _init_completion >/dev/null 2>&1; then")
    script.append("    _init_completion || return")
    script.append("    else")
    script.append("        COMPREPLY=()")
    script.append("        cur=\"${COMP_WORDS[COMP_CWORD]}\"")
    script.append("        prev=\"${COMP_WORDS[COMP_CWORD-1]}\"")
    script.append("        words=(\"${COMP_WORDS[@]}\")")
    script.append("        cword=$COMP_CWORD")
    script.append("    fi")
    script.append("")
    script.append("    local command_chain")
    script.append("    command_chain=()")
    script.append("")
    script.append("    # Find the command chain")
    script.append("    for ((i=1; i < cword; i++)); do")
    script.append("        if [[ \"${words[i]}\" != -* ]]; then")
    script.append("            command_chain+=(\"${words[i]}\")")
    script.append("        fi")
    script.append("    done")
    script.append("")
    script.append("    local cmd_str=\"${command_chain[*]}\"")
    script.append("")
    script.append("    case \"$cmd_str\" in")

    # Helper to flatten options
    def get_opts(node):
        opts = node.get("options", {})
        if isinstance(opts, dict):
            return " ".join(sorted(opts.keys()))
        return " ".join(sorted(list(set(opts))))

    # Helper to get subcommands
    def get_subs(node):
        return " ".join(sorted(node.get("subcommands", {}).keys()))

    # Recursive function to generate cases
    def visit(node, path):
        path_str = " ".join(path)
        opts = get_opts(node)
        subs = get_subs(node)

        # Case for this path
        script.append(f"        \"{path_str}\")")
        if subs:
            script.append(f"            COMPREPLY=( $(compgen -W \"{subs} {opts}\" -- \"$cur\") )")
        else:
            script.append(f"            COMPREPLY=( $(compgen -W \"{opts}\" -- \"$cur\") )")
        script.append("            return 0")
        script.append("            ;;")

        for sub_name, sub_node in node.get("subcommands", {}).items():
            visit(sub_node, path + [sub_name])

    visit(tree, [])

    script.append("    esac")
    script.append("}")
    script.append("complete -F _kicad_cli kicad-cli")

    return "\n".join(script)

if __name__ == "__main__":
    tree = explore([])
    print(generate_bash_completion(tree))



from conans import ConanFile

class ArgparseConan(ConanFile):
    name = "argparse"
    version = "3.0"
    exports_sources = "include/argparse.hpp"
    no_copy_source = True

    def package(self):
        self.copy("argparse.hpp")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#####################################
#
# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright (C) 2016 Nick Østergaard <oe.nick at gmail dot com>
# Copyright (C) 2016 KiCad Developers
#
# License GNU GPL Version 3 or any later version.
#
#####################################

import matplotlib.pyplot as plt
import csv
import numpy as np
import time

# Initialize data structure variable
data = []

# Read CSV file
with open('i18n_status.csv', 'r') as csvfile:
    spamreader = csv.reader(csvfile, delimiter=';')
    for row in spamreader:
        data.append(row)

# Replace empyt values with zero and convert numbers to int
for index,value in np.ndenumerate( data ):
    if value=='':
        data[index[0]][index[1]] = 0
    try:
        data[index[0]][index[1]] = int(data[index[0]][index[1]])
    except:
        pass

# Sort data after mostly translated
data[1:] = sorted(data[1:], key=lambda x: int(x[1]), reverse=True)

# Prepare some number for formatting the plot
N = len(data[1:])          # number of languages
width = 0.35               # the width of the bars
ind = np.arange(N)*width*5 # the x locations for the groups

# Plot the bars
fig, ax = plt.subplots()
rects1 = ax.bar(ind, list(zip(*data))[1][1:], width, color='b')
rects2 = ax.bar(ind+width, list(zip(*data))[2][1:], width, color='r')
rects3 = ax.bar(ind+2*width, list(zip(*data))[3][1:], width, color='y')

# Plot ceiling
max_nof_strings = sum(map(int, data[1][1:4]))
plt.plot([0,max(ind)+3*width],[max_nof_strings,max_nof_strings], color='k', linewidth='2')
ax.set_xlim([0,max(ind)+3*width])

# Add some text for labels, title and axes ticks
ax.set_ylabel('Number of strings')
ax.set_title('Translation status')
ax.set_xticks(ind+width*1.5)
ax.set_xticklabels(list(zip(*data))[0][1:])
ax.yaxis.grid(True, which='both') # horizontal lines
ax.legend((rects1[0], rects2[0], rects3[0]), ('TRANSLATED', 'FUZZY', 'UNTRANSLATED'), loc='upper center', bbox_to_anchor=(0.5, -0.05), fancybox=True, ncol=3)
plt.figtext(0.99, 0.96, time.strftime("%d/%m %Y"), horizontalalignment='right')
plt.subplots_adjust(left=0.07, right=0.99, top=0.95, bottom=0.12)

fig.set_size_inches(12, 8)

fig.savefig('i18n_status.svg')

# Show the magic to the user
plt.show()



#
# Example python script to generate a BOM from a KiCad generic netlist
#

"""
    @package
    Output: Cadstar RINF netlist
    Sorted By: Ref

    Command line:
    python "pathToFile/netlist_form_cadstar.py" "%I" "%O.frp"
"""

from __future__ import print_function

# Import the KiCad python helper module
import kicad_netlist_reader
import sys

# A helper function to convert a UTF8/Unicode/locale string read in netlist
# for python2 or python3 (Windows/unix)
def fromNetlistText( aText ):
    if sys.platform.startswith('win32'):
        try:
            return aText.encode('utf-8').decode('cp1252')
        except UnicodeDecodeError:
            return aText
    else:
        return aText

# Generate an instance of a generic netlist, and load the netlist tree from
# the command line option. If the file doesn't exist, execution will stop
netlist = kicad_netlist_reader.netlist(sys.argv[1])

# Open a file to write to, if the file cannot be opened output to stdout
# instead
canOpenFile = True
try:
    f = open(sys.argv[2], 'wb')
except IOError:
    e = "Can't open output file for writing: " + sys.argv[2]
    print(__file__, ":", e, sys.stderr)
    f = sys.stdout
    canOpenFile = False

components = netlist.getInterestingComponents( excludeBoard=True )

row =""

''' Netlist header '''
row += ".HEA" + '\n'
''' Generate line .TIM <time> '''
row += '.TIM ' + netlist.getDate() + '\n'
''' Generate line .APP <eeschema version> '''
row += '.APP ' + netlist.getTool() + '\n'
row += '.TYP FULL' + '\n\n'


''' Generate list of component
 for each component  create lines like
    .ADD_COM U3 "74LS541"   (when no footprint name specified)
    .ADD_COM JP1 "CONN_8X2" "pin_array_8x2"   (with a specified footprint name)
'''

for c in components:
    c.getExcludeFromBOM
    row += ".ADD_COM " + " " + c.getRef() + " \"" + c.getValue() + "\""

    fp_name = c.getFootprint( False )

    if fp_name != "":
        row += " \"" + fp_name + "\""

    row += '\n'

'''
generate for each net create lines like
.ADD_TER U3.9 "/PC-RST"
.TER     U3.8
         BUS1.2
.ADD_TER BUS1.14 "/PC-IOR"
.TER     U3.7
'''
nets = netlist.getNets()

row += '\n'

for net in nets:
    # count the number of pads in net. nets with only one pad are skipped
    netitems = net.children
    pad_count = 0

    for node in netitems:
        pad_count += 1

    item_cnt = 1
    netitems = net.children

    if pad_count > 1:
        for node in netitems:

            if item_cnt == 1:
                row += ".ADD_TER " + net.get( "node", "ref" ) + '.' + net.get( "node", "pin" )
                row += " \"" + net.get( "net", "name" ) + '\"\n'

            if item_cnt == 2:
                row += ".TER     "

            if item_cnt > 2:
                row += "         "

            if item_cnt > 1:
                    row += node.get('node','ref') + '.' + node.get('node','pin') + '\n'

            item_cnt += 1


row += '\n.END\n'

f.write(row.encode('utf-8'))
f.close()

if not canOpenFile:
    sys.exit(1)


#
# This program source code file is part of KiCad, a free EDA CAD application.
#
# Copyright (C) 2021 KiCad Developers, see AUTHORS.txt for contributors.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

from __future__ import division
import pcbnew
import math


class FootprintWizard(pcbnew.FootprintWizardPlugin):
    """!
    A class to simplify many aspects of footprint creation, leaving only
    the footprint specific routines to the wizards themselves.

    Inherit this class to make a new wizard.

    Provides simplified access to helpers like drawing functions, a transform
    matrix stack and simple parameter checking.

    Generally, you need to implement:
        GetValue()
        GenerateParameterList()
        CheckParameters()
        BuildThisFootprint()
        GetName()
        GetDescription()
    """

    # Copy units from pcbnew
    uMM = pcbnew.uMM
    uMils = pcbnew.uMils
    uFloat = pcbnew.uFloat
    uInteger = pcbnew.uInteger
    uBool = pcbnew.uBool
    uRadians = pcbnew.uRadians
    uDegrees = pcbnew.uDegrees
    uPercent = pcbnew.uPercent
    uString = pcbnew.uString

    def __init__(self):
        pcbnew.FootprintWizardPlugin.__init__(self)
        self.GenerateParameterList()

    def GetName(self):
        """!
        Return the name of the footprint wizard
        """
        raise NotImplementedError

    def GetDescription(self):
        """!
        Return the footprint wizard description
        """
        raise NotImplementedError

    def GetValue(self):
        """!
        Return the value (name) of the generated footprint
        """
        raise NotImplementedError

    def GenerateParameterList(self):
        """!
        Footprint parameter specification is done here
        """
        raise NotImplementedError

    def CheckParameters(self):
        """!
        Any custom parameter checking should be performed here
        """
        raise NotImplementedError

    def BuildThisFootprint(self):
        """!
        Draw the footprint.

        This is specific to each footprint class, you need to implement
        this to draw what you want
        """
        raise NotImplementedError

    # Do not override this method!
    def BuildFootprint(self):
        """!
        Actually make the footprint. We defer all but the set-up to
        the implementing class
        """

        self.buildmessages = ""
        self.module = pcbnew.FOOTPRINT(None)  # create a new module

        # Perform default checks on all parameters
        for p in self.params:
            p.ClearErrors()
            p.Check()  # use defaults

        self.CheckParameters()  # User error checks

        if self.AnyErrors():  # Errors were detected!

            self.buildmessages = ("Cannot build footprint: "
                                  "Parameters have errors:\n")

            for p in self.params:
                if len(p.error_list) > 0:
                    self.buildmessages += "['{page}']['{name}']:\n".format(
                        page=p.page, name=p.name)

                    for error in p.error_list:
                        self.buildmessages += "\t" + error + "\n"

            return

        self.buildmessages = (
            "Building new {name} footprint with the following parameters:\n"
            .format(name=self.name))

        self.buildmessages += self.Show()

        self.draw = FootprintWizardDrawingAids(
            self.module)

        self.module.SetValue(self.GetValue())
        self.module.SetReference("%s**" % self.GetReferencePrefix())

        fpid = pcbnew.LIB_ID("", self.module.GetValue())  # the lib name  (empty) and the name in library
        self.module.SetFPID(fpid)

        self.SetModule3DModel()  # add a 3D module if specified

        thick = self.GetTextThickness()

        self.module.Reference().SetTextThickness(thick)
        self.module.Value().SetTextThickness(thick)

        self.BuildThisFootprint()  # implementer's build function

        return

    def SetModule3DModel(self):
        """!
        If your plug-in sets a 3D model, override this function
        """
        pass

    def GetTextSize(self):
        """!
        Get the default text size for the footprint. Override to change it.

        Defaults to IPC nominal of 1.0mm
        """
        return pcbnew.FromMM(1.0)

    def GetTextThickness(self):
        """!
        Thicker than IPC guidelines (10% of text height = 0.12mm)
        as 5 wires/mm is a common silk screen limitation
        """
        return pcbnew.FromMM(0.15)


class FootprintWizardDrawingAids:
    """!
    Collection of handy functions to simplify drawing shapes from within
    footprint wizards

    A "drawing context" is provided which can be used to set and retain
    settings such as line thickness and layer. The DC also contains a
    "transform stack", which allows easy positioning and transforming of
    drawn elements without lots of geometric book-keeping.
    """

    # directions (in degrees, compass-like)
    dirN = 0
    dirNE = 45
    dirE = 90
    dirSE = 135
    dirS = 180
    dirSW = 225
    dirW = 270
    dirNW = 315

    # Flip constants
    flipNone = 0  # no flip transform
    flipX = 1  # flip X values, i.e. about the Y-axis
    flipY = 2  # flip Y values, i.e. about the X-axis
    flipBoth = 3  # flip X and Y values, equivalent to a 180-degree rotation

    xfrmIDENTITY = [1, 0, 0, 0, 1, 0]  # no transform

    # these values come from our KiCad Library Convention 0.11
    defaultLineThickness = pcbnew.FromMM(0.15)

    def DefaultGraphicLayer(self):
        return pcbnew.F_SilkS

    def DefaultTextValueLayer(self):
        return pcbnew.F_Fab

    def __init__(self, module):
        self.module = module
        # drawing context defaults
        self.dc = {
            'layer': self.DefaultGraphicLayer(),
            'lineThickness': self.defaultLineThickness,
            'transforms': [],
            'transform': self.xfrmIDENTITY
        }

    def PushTransform(self, mat):
        """!
        Add a transform to the top of the stack and recompute the
        overall transform

        @param mat: the transform matrix to add to the stack
        """
        self.dc['transforms'].append(mat)
        self.RecomputeTransforms()

    def PopTransform(self, num=1):
        """!
        Remove a transform from the top of the stack and recompute the
        overall transform

        @param num: the number of transforms to pop from the stack.
        @return the last popped transform
        """

        for i in range(num):
            mat = self.dc['transforms'].pop()
        self.RecomputeTransforms()
        return mat

    def ResetTransform(self):
        """!
        Reset the transform stack to the identity matrix.
        """
        self.dc['transforms'] = []
        self.RecomputeTransforms()

    def _ComposeMatricesWithIdentity(self, mats):
        """!
        Compose a sequence of matrices together by sequential
        pre-multiplication with the identity matrix.

        @param mats: list of matrices to compose
        @return: the composed transform matrix
        """

        x = self.xfrmIDENTITY

        for mat in mats:
            # Pre-compose with each transform in turn
            x = [
                x[0] * mat[0] + x[1] * mat[3],
                x[0] * mat[1] + x[1] * mat[4],
                x[0] * mat[2] + x[1] * mat[5] + x[2],
                x[3] * mat[0] + x[4] * mat[3],
                x[3] * mat[1] + x[4] * mat[4],
                x[3] * mat[2] + x[4] * mat[5] + x[5]]

        return x

    def RecomputeTransforms(self):
        """!
        Re-compute the transform stack into a single transform and
        store in the DC
        """
        self.dc['transform'] = self._ComposeMatricesWithIdentity(
            self.dc['transforms'])

    def TransformTranslate(self, x, y, push=True):
        """!
        Set up and return a transform matrix representing a translation
        optionally pushing onto the stack

        (   1  0   x  )
        (   0  1   y  )

        @param x: translation in x-direction
        @param y: translation in y-direction
        @param push: add this transform to the current stack
        @return the generated transform matrix
        """
        mat = [1, 0, x, 0, 1, y]

        if push:
            self.PushTransform(mat)
        return mat

    def TransformFlipOrigin(self, flip, push=True):
        """!
        Set up and return a transform matrix representing a horizontal,
        vertical or both flip about the origin

        @param flip: one of flipNone, flipX, flipY, flipBoth
        @param push: add this transform to the current stack
        @return the generated transform matrix
        """
        mat = None
        if flip == self.flipX:
            mat = [-1, 0, 0, 0, 1, 0]
        elif flip == self.flipY:
            mat = [1, 0, 0, 0, -1, 0]
        elif flip == self.flipBoth:
            mat = [-1, 0, 0, 0, -1, 0]
        elif flip == self.flipNone:
            mat = self.xfrmIDENTITY
        else:
            raise ValueError

        if push:
            self.PushTransform(mat)
        return mat

    def TransformFlip(self, x, y, flip=flipNone, push=True):
        """!
        Set up and return a transform matrix representing a horizontal,
        vertical or both flip about a point (x,y)

        This is performed by a translate-to-origin, flip, translate-
        back sequence.

        @param x: the x coordinate of the flip point
        @param y: the y coordinate of the flip point
        @param flip: one of flipNone, flipX, flipY, flipBoth
        @param push: add this transform to the current stack
        @return the generated transform matrix
        """
        mats = [self.TransformTranslate(x, y, push=False),
                self.TransformFlipOrigin(flip, push=False),
                self.TransformTranslate(-x, -y, push=False)]

        # Distil into a single matrix
        mat = self._ComposeMatricesWithIdentity(mats)

        if push:
            self.PushTransform(mat)
        return mat

    def TransformRotationOrigin(self, rot, push=True):
        """!
        Set up and return a transform matrix representing a rotation
        about the origin, and optionally push onto the stack

        (   cos(t)  -sin(t)   0  )
        (   sin(t)   cos(t)   0  )

        @param rot: the rotation angle in degrees
        @param push: add this transform to the current stack
        @return the generated transform matrix
        """
        rads = rot * math.pi / 180
        mat = [math.cos(rads), -math.sin(rads), 0,
               math.sin(rads), math.cos(rads), 0]

        if push:
            self.PushTransform(mat)
        return mat

    def TransformRotation(self, x, y, rot, push=True):
        """!
        Set up and return a transform matrix representing a rotation
        about the point (x,y), and optionally push onto the stack

        This is performed by a translate-to-origin, rotate, translate-
        back sequence

        @param x: the x coordinate of the rotation centre
        @param y: the y coordinate of the rotation centre
        @param rot: the rotation angle in degrees
        @param push: add this transform to the current stack
        @return the generated transform matrix
        """

        mats = [self.TransformTranslate(x, y, push=False),
                self.TransformRotationOrigin(rot, push=False),
                self.TransformTranslate(-x, -y, push=False)]

        # Distil into a single matrix
        mat = self._ComposeMatricesWithIdentity(mats)

        if push:
            self.PushTransform(mat)
        return mat

    def TransformScaleOrigin(self, sx, sy=None, push=True):
        """!
        Set up and return a transform matrix representing a scale about
        the origin, and optionally push onto the stack

        (   sx   0   0  )
        (    0  sy   0  )

        @param sx: the scale factor in the x direction
        @param sy: the scale factor in the y direction
        @param push: add this transform to the current stack
        @return the generated transform matrix
        """

        if sy is None:
            sy = sx

        mat = [sx, 0, 0, 0, sy, 0]

        if push:
            self.PushTransform(mat)
        return mat

    def TransformPoint(self, x, y, mat=None):
        """!
        Return a point (x, y) transformed by the given matrix, or if
        that is not given, the drawing context transform

        @param x: the x coordinate of the point to transform
        @param y: the y coordinate of the point to transform
        @param mat: the transform matrix to use or None to use the current DC's
        @return: the transformed point as a VECTOR2I
        """

        if not mat:
            mat = self.dc['transform']

        return pcbnew.VECTOR2I( (int)(x * mat[0] + y * mat[1] + mat[2]),
                                (int)(x * mat[3] + y * mat[4] + mat[5]) )

    def SetLineThickness(self, lineThickness):
        """!
        Set the current pen lineThickness used for subsequent drawing
        operations

        @param lineThickness: the new line thickness to set
        """
        self.dc['lineThickness'] = lineThickness

    def SetLineTickness(self, lineThickness):
        """!
        Old version of SetLineThickness.
        Does the same thing, but is only here for compatibility with old
        scripts.
        Set the current pen lineThickness used for subsequent drawing
        operations

        @param lineThickness: the new line thickness to set
        """
        self.SetLineThickness(lineThickness)

    def GetLineThickness(self):
        """!
        Get the current drawing context line thickness
        """
        return self.dc['lineThickness']

    def SetLayer(self, layer):
        """!
        Set the current drawing layer, used for subsequent drawing
        operations
        """
        self.dc['layer'] = layer

    def GetLayer(self):
        """!
        Return the current drawing layer, used for drawing operations
        """
        return self.dc['layer']

    def Line(self, x1, y1, x2, y2):
        """!
        Draw a line from (x1, y1) to (x2, y2)
        """
        outline = pcbnew.PCB_SHAPE(self.module)
        outline.SetWidth(self.GetLineThickness())
        outline.SetLayer(self.GetLayer())
        outline.SetShape(pcbnew.S_SEGMENT)
        start = self.TransformPoint(x1, y1)
        end = self.TransformPoint(x2, y2)
        outline.SetStartEnd(start, end)
        self.module.Add(outline)

    def Circle(self, x, y, r, filled=False):
        """!
        Draw a circle at (x,y) of radius r
        If filled is true, the thickness and radius of the line will be set
        such that the circle appears filled

        @param x: the x coordinate of the arc centre
        @param y: the y coordinate of the arc centre
        @param r: the circle's radius
        @param filled: True to draw a filled circle, False to use the current
                       DC line thickness
        """

        circle = pcbnew.PCB_SHAPE(self.module)
        start = self.TransformPoint(x, y)

        if filled:
            circle.SetWidth(r)
            end = self.TransformPoint(x, y + r/2)
        else:
            circle.SetWidth(self.dc['lineThickness'])
            end = self.TransformPoint(x, y + r)

        circle.SetLayer(self.dc['layer'])
        circle.SetShape(pcbnew.S_CIRCLE)
        circle.SetStartEnd(start, end)
        self.module.Add(circle)

    def MyCmp(self, n1, n2):
        """
        replace the cmp() of python2
        """
        if n1 < n2:
            return -1
        if n1 > n2:
            return 1
        return 0

    def Arc(self, cx, cy, sx, sy, angle):
        """!
        Draw an arc based on centre, start and angle

        The transform matrix is applied

        Note that this won't work properly if the result is not a
        circular arc (e.g. a horizontal scale)

        @param cx: the x coordinate of the arc centre
        @param cy: the y coordinate of the arc centre
        @param sx: the x coordinate of the arc start point
        @param sy: the y coordinate of the arc start point
        @param angle: the arc's central angle (in deci-degrees)
        """
        arc = pcbnew.PCB_SHAPE(self.module)
        arc.SetShape(pcbnew.SHAPE_T_ARC)
        arc.SetWidth(self.dc['lineThickness'])

        center = self.TransformPoint(cx, cy)
        start = self.TransformPoint(sx, sy)

        arc.SetLayer(self.dc['layer'])

        # check if the angle needs to be reverse (a flip scaling)
        if self.MyCmp(self.dc['transform'][0], 0) != self.MyCmp(self.dc['transform'][4], 0):
            angle = -angle

        arc.SetCenter(center)
        arc.SetStart(start)
        arc.SetArcAngleAndEnd(angle, True)
        #arc.SetLocalCoord()
        self.module.Add(arc)

    def HLine(self, x, y, l):
        """!
        Draw a horizontal line from (x,y), rightwards

        @param x: line start x coordinate
        @param y: line start y coordinate
        @param l: line length
        """
        self.Line(x, y, x + l, y)

    def VLine(self, x, y, l):
        """!
        Draw a vertical line from (x1,y1), downwards

        @param x: line start x coordinate
        @param y: line start y coordinate
        @param l: line length
        """
        self.Line(x, y, x, y + l)

    def Polyline(self, pts, mirrorX=None, mirrorY=None):
        """!
        Draw a polyline, optionally mirroring around the given points

        @param pts: list of polyline vertices (list of (x, y))
        @param mirrorX: x coordinate of mirror point (None for no x-flip)
        @param mirrorY: y coordinate of mirror point (None for no y-flip)
        """

        def _PolyLineInternal(pts):
            if len(pts) < 2:
                return

            for i in range(0, len(pts) - 1):
                self.Line(pts[i][0], pts[i][1], pts[i+1][0], pts[i+1][1])

        if mirrorX is None and mirrorY is None:
            _PolyLineInternal(pts)  # original

        elif mirrorX is not None and mirrorY is not None:
            self.TransformFlip(mirrorX, mirrorY, self.flipBoth)  # both
            _PolyLineInternal(pts)
            self.PopTransform()

        elif mirrorX is not None:
            self.TransformFlip(mirrorX, 0, self.flipX)
            _PolyLineInternal(pts)
            self.PopTransform()

        elif mirrorY is not None:
            self.TransformFlip(0, mirrorY, self.flipY)
            _PolyLineInternal(pts)
            self.PopTransform()

    def Reference(self, x, y, size, orientation_degree=0):
        """!
        Draw the module's reference as the given point.

        The actual setting of the reference is not done in this drawing
        aid - that is up to the wizard

        @param x: the x position of the reference
        @param y: the y position of the reference
        @param size: the text size (in both directions)
        @param orientation_degree: text orientation in degrees
        """

        text_size = pcbnew.VECTOR2I(size, size)

        self.module.Reference().SetPosition( self.TransformPoint(x, y) )
        self.module.Reference().SetTextSize(text_size)
        self.module.Reference().SetTextAngle( pcbnew.EDA_ANGLE( orientation_degree, pcbnew.DEGREES_T ) )

    def Value(self, x, y, size, orientation_degree=0):
        """!
        As for references, draw the module's value

        @param x: the x position of the value
        @param y: the y position of the value
        @param size: the text size (in both directions)
        @param orientation_degree: text orientation in degrees
        """
        text_size = pcbnew.VECTOR2I(size, size)

        self.module.Value().SetPosition(self.TransformPoint(x, y))
        self.module.Value().SetTextSize(text_size)
        self.module.Value().SetLayer(self.DefaultTextValueLayer())
        self.module.Value().SetTextAngle( pcbnew.EDA_ANGLE( orientation_degree, pcbnew.DEGREES_T ) )

    def Box(self, x, y, w, h):
        """!
        Draw a rectangular box, centred at (x,y), with given width and
        height

        @param x: the x coordinate of the box's centre
        @param y: the y coordinate of the box's centre
        @param w: the width of the box
        @param h: the height of the box
        """

        pts = [[x - w/2, y - h/2],  # left
               [x + w/2, y - h/2],  # right
               [x + w/2, y + h/2],  # bottom
               [x - w/2, y + h/2],  # top
               [x - w/2, y - h/2]]  # close

        self.Polyline(pts)

    def NotchedCircle(self, x, y, r, notch_w, notch_h, rotate=0):
        """!
        Circle radius r centred at (x, y) with a raised or depressed notch
        at the top
        Notch height is measured from the top of the circle radius

        @param x: the x coordinate of the circle's centre
        @param y: the y coordinate of the circle's centre
        @param r: the radius of the circle
        @param notch_w: the width of the notch
        @param notch_h: the height of the notch
        @param rotate: the rotation of the whole figure, in degrees
        """

        self.TransformRotation(x, y, rotate)

        # find the angle where the notch vertical meets the circle
        angle_intercept = math.asin(notch_w/(2 * r))

        # and find the co-ords of this point
        sx = math.sin(angle_intercept) * r
        sy = -math.cos(angle_intercept) * r

        # NOTE: this may be out by a factor of ten one day
        arc_angle = (math.pi * 2 - angle_intercept * 2) * (1800/math.pi)

        self.Arc(x, y, sx, sy, arc_angle)

        pts = [[sx,  sy],
               [sx,  -r - notch_h],
               [-sx, -r - notch_h],
               [-sx, sy]]

        self.Polyline(pts)
        self.PopTransform()

    def NotchedBox(self, x, y, w, h, notchW, notchH, rotate=0):
        """!
        Draw a box with a notch in the centre of the top edge

        @param x: the x coordinate of the circle's centre
        @param y: the y coordinate of the circle's centre
        @param w: the width of the box
        @param h: the height of the box
        @param notchW: the width of the notch
        @param notchH: the height of the notch
        @param rotate: the rotation of the whole figure, in degrees
        """

        self.TransformRotation(x, y, rotate)

        # limit to half the overall width
        notchW = min(x + w/2, notchW)

        # draw notch
        self.Polyline([  # three sides of box
            (x - w/2, y - h/2),
            (x - w/2, y + h/2),
            (x + w/2, y + h/2),
            (x + w/2, y - h/2),
            # the notch
            (notchW/2, y - h/2),
            (notchW/2, y - h/2 + notchH),
            (-notchW/2, y - h/2 + notchH),
            (-notchW/2, y - h/2),
            (x - w/2, y - h/2)
        ])

        self.PopTransform()

    def BoxWithDiagonalAtCorner(self, x, y, w, h,
                                setback=pcbnew.FromMM(1.27), flip=flipNone):
        """!
        Draw a box with a diagonal at the top left corner.

        @param x: the x coordinate of the circle's centre
        @param y: the y coordinate of the circle's centre
        @param w: the width of the box
        @param h: the height of the box
        @param setback: the set-back of the diagonal, in both x and y
        @param flip: one of flipNone, flipX, flipY or flipBoth to change the
         diagonal corner
        """

        self.TransformFlip(x, y, flip, push=True)

        pts = [[x - w/2 + setback, y - h/2],
               [x - w/2,           y - h/2 + setback],
               [x - w/2,           y + h/2],
               [x + w/2,           y + h/2],
               [x + w/2,           y - h/2],
               [x - w/2 + setback, y - h/2]]

        self.Polyline(pts)

        self.PopTransform()

    def BoxWithOpenCorner(self, x, y, w, h,
                          setback=pcbnew.FromMM(1.27), flip=flipNone):
        """!
        Draw a box with an opening at the top left corner

        @param x: the x coordinate of the circle's centre
        @param y: the y coordinate of the circle's centre
        @param w: the width of the box
        @param h: the height of the box
        @param setback: the set-back of the opening, in both x and y
        @param flip: one of flipNone, flipX, flipY or flipBoth to change the
         open corner position
        """

        self.TransformTranslate(x, y)
        self.TransformFlipOrigin(flip)

        pts = [[- w/2,           - h/2 + setback],
               [- w/2,           + h/2],
               [+ w/2,           + h/2],
               [+ w/2,           - h/2],
               [- w/2 + setback, - h/2]]

        self.Polyline(pts)

        self.PopTransform(num=2)

    def RoundedBox(self, x, y, w, h, rad):
        """!
        Draw a box with rounded corners (i.e. a 90-degree circular arc)

        :param x: the x coordinate of the box's centre
        :param y: the y coordinate of the box's centre
        :param w: the width of the box
        :param h: the height of the box
        :param rad: the radius of the corner rounds
        """

        x_inner = w - rad * 2
        y_inner = h - rad * 2

        x_left = x - w / 2
        y_top = y - h / 2

        # Draw straight sections
        self.HLine(x_left + rad, y_top, x_inner)
        self.HLine(x_left + rad, -y_top, x_inner)

        self.VLine(x_left, y_top + rad, y_inner)
        self.VLine(-x_left, y_top + rad, y_inner)

        # corner arcs
        ninety_deg = 90 * 10  # deci-degs
        cx = x - w / 2 + rad
        cy = y - h / 2 + rad

        # top left
        self.Arc(+cx, +cy, +x_left, +cy, +ninety_deg)
        self.Arc(-cx, +cy, -x_left, +cy, -ninety_deg)
        self.Arc(+cx, -cy, +x_left, -cy, -ninety_deg)
        self.Arc(-cx, -cy, -x_left, -cy, +ninety_deg)

    def ChamferedBox(self, x, y, w, h, chamfer_x, chamfer_y):
        """!
        Draw a box with chamfered corners.

        :param x: the x coordinate of the box's centre
        :param y: the y coordinate of the box's centre
        :param w: the width of the box
        :param h: the height of the box
        :param chamfer_x: the size of the chamfer set-back in the x direction
        :param chamfer_y: the size of the chamfer set-back in the y direction
        """
        # outermost dimensions
        x_left = x - w / 2
        y_top = y - h / 2

        # x and y coordinates of inner edges of chamfers
        x_inner = x_left + chamfer_x
        y_inner = y_top + chamfer_y

        pts = [
            [+x_inner, +y_top],
            [-x_inner, +y_top],
            [-x_left,  +y_inner],
            [-x_left,  -y_inner],
            [-x_inner, -y_top],
            [+x_inner, -y_top],
            [+x_left,  -y_inner],
            [+x_left,  +y_inner],
            [+x_inner, +y_top],
        ]

        self.draw.Polyline(pts)

    def MarkerArrow(self, x, y, direction=dirN, width=pcbnew.FromMM(1)):
        """!
        Draw a marker arrow facing in the given direction, with the
        point at (x,y)

        @param x: x position of the arrow tip
        @param y: y position of the arrow tip
        @param direction: arrow direction in degrees (0 is "north", can use
         dir* shorthands)
        @param width: arrow width
        """

        self.TransformTranslate(x, y)
        self.TransformRotationOrigin(direction)

        pts = [[0,          0],
               [width / 2,  width / 2],
               [-width / 2, width / 2],
               [0,          0]]

        self.Polyline(pts)
        self.PopTransform(2)
